;; Problem 162
;; 05 October 2007
;;
;; In the hexadecimal number system numbers are represented using 16
;; different digits:
;;
;; 0,1,2,3,4,5,6,7,8,9,A,B,C,D,E,F
;;
;; The hexadecimal number AF when written in the decimal number system
;; equals 10x16+15=175.
;;
;; In the 3-digit hexadecimal numbers 10A, 1A0, A10, and A01 the
;; digits 0,1 and A are all present.
;; Like numbers written in base ten we write hexadecimal numbers without leading zeroes.
;;
;; How many hexadecimal numbers containing at most sixteen hexadecimal
;; digits exist with all of the digits 0,1, and A present at least
;; once?  Give your answer as a hexadecimal number.
;;
;; (A,B,C,D,E and F in upper case, without any leading or trailing
;; code that marks the number as hexadecimal and without leading
;; zeroes , e.g. 1A3F and not: 1a3f and not 0x1a3f and not $1A3F and
;; not #1A3F and not 0000001A3F)

(defmacro dbg (&rest rest)
  nil)

(defun meets-condition (str)
  (declare (type string str))
  (if
   (and
    (not (equal "0" (subseq str 0 1)))
    (find #\A str)
    (find #\0 str)
    (find #\1 str))
   t
   nil))

(assert (not (meets-condition "01A")))
(assert      (meets-condition "10A"))
(assert      (meets-condition "FFFFFF10A"))

(defmacro acollect (&body body)
  (let ((resv (gensym)))
    `(let ((it (list))
	   (,resv nil))
      (setf ,resv (progn ,@body))
      (values it ,resv))))

(defun enum-naieve (ndigits)
  (acollect 
    (loop for ii from 0 to (- (expt 16 ndigits) 1)
	  for str = (format nil "~x" ii) 
               then (format nil "~x" ii) 
	  do
	  (when (meets-condition str)
	    (push str it))
 	  (dbg
	   (format t "ndigits:~a ii:~a 0x~a~&"
		   ndigits
		   ii
		   str)))))

(defmacro aformat (expr fmt &rest args)
  `(let ((it ,expr))
    (format t ,fmt ,@args)
    it))

(aformat (enum-naieve 3) "length:~a~&" (length it) it)
; (aformat (enum-naieve 4) "length:~a~&" (length it) it)

(loop for ii from 1 to 5
      do
      (format t "length(~a):~a~&" ii (length (enum-naieve ii))))

;; length(1):0
;; length(2):0
;; length(3):4
;; length(4):262
;; length(5):10190

(defun bad-count-hex-combinations (acc ndigits)
  ;; add together the following counts:
  ;;  - fixing the first digit at '1', 2 at 'A', 3 at '0' then
  ;;    (expt 16 (- ndigits 3))
  ;;  - fixing the first digit at '1', 2 at '0', 3 at 'A' then 
  ;;    (expt 16 (- ndigits 3))
  ;;  - fixing the first digit at 'A', 2 at '1', 3 at '0' then 
  ;;    (expt 16 (- ndigits 3))
  ;;  - fixing the first digit at 'A', 2 at '0', 3 at '1' then 
  ;;    (expt 16 (- ndigits 3))

  ;; ** this is (* 4 (expt (ndigits 3 ))) since we 'fixed' 3 of the
  ;; locations.
  ;;
  ;;  - then the first digit can be any of the other remaining 13 digits
  ;;    so it is 13 times the following:
  ;;    fix the next 3 digits as '1','A', and '0'
  ;;    then (expt 16 (- ndigits 4))
  ;; this gets us 'count'
  ;; if ndigits < 3 just stop
  ;; otherwise recurse with (count-hex-combinations (+ count acc) (- ndigits 1))
  (cond ((< ndigits 3)
	 acc)
	((= ndigits 3)
	 (+ acc 4))
	(t
	 (let ((count 0))
	   (setf count (+ count (* 4 (expt 16 (- ndigits 3)))))
;;	   (if (> ndigits 4)
;;	       (setf count (+ count (* 13 (expt 16 (- ndigits 4))))))
	   (setf count (+ count (* 13 (expt 16 (- ndigits 4)))))
	   (count-hex-combinations
	    (+ count acc)
	    (- ndigits 1))))))

;(count-hex-combinations 0 3) ; 4
;(count-hex-combinations 0 4) ; 81


;(loop for str in (enum-naieve 4)
;      do
;      (format t "~a~&" str))


(defmacro aif (test consequent &optional otherwise)
  `(let ((it ,test))
    (if it
	,consequent
	,otherwise)))

(defun string-starts-wtih (str pat)
  (aif (search str pat)
      (= 0 it)
      nil))

;; (let ((stuff (enum-naieve 4)))
;;   (apply #'+
;; 	 (mapcar #'(lambda (chr)
;; 		     (length (remove-if-not #'(lambda (elt) (string-starts-wtih chr elt))
;; 					    stuff)))
;; 		 '("2" "3" "4" "5" "6" "7" "8" "9" "B" "C" "D" "E" "F"))))
;; => 78


;; (let ((stuff (enum-naieve 4)))
;;   (apply #'+
;; 	 (mapcar #'(lambda (chr)
;; 		     (length (remove-if-not #'(lambda (elt) (string-starts-wtih chr elt))
;; 					    stuff)))
;; 		 '("A" "1"))))
;; => 184

'(with-open-file (out "/home/mortis/personal/projects/project-euler/x.x" :direction :output :if-exists :supersede)
  (loop for str in 
   (remove-if-not #'(lambda (str) (and (= 4 (length str)) (string-starts-wtih "1" str)))
    (sort (enum-naieve 4) #'string<))
   do
   (format out "~a~&" str)))

(defun grep-str-starts-with (str items)
  (remove-if-not #'(lambda (item)
		     (string-starts-wtih str item))
		 items))


;; when len = 3 there are 4 perms
;;  A01 A10 10A 1A0
;; when len = 4, and the first char is one of (2,3,4,5,6,7,8,9,B,C,D,E,F)
;;     there are 6 permutations of the 3 necessary chars (3! of 0,1,A)
;;     or 13 * 3!
;;   when the first char is 1, or A, there are 90 permutations? -- is this b/c 6 are already covered?
;;    1A0* = 16
;;    10A* = 16
;;    1A*0 = 16
;;    10*A = 16
;;    1*A0 = 16
;;    1*0A = 16
;;  yes, it is b/c 6 cases are duplicately covered:
;;    1A0A was already counted by the 16 of 1A0*
;;    which 'overlaps' w/the count from 1*0A
;;  so we have to subtract off fact(ndigits-1)
;;  This is A, 0, and * (anything), or again, 3!, or 6 permutations where 1 is at the front,
;;  the twist is that the * (anything) slot is actually 16 things
;;  Since the A can also be at the front, we have 2 * 16 * 3!

;; this expression is an embodiement of the above logic for a 4 digit
;; hex number:
(+
 (+ (* 13 (fact 3))                     ;; the starting digits other than 1/A
    (* 2 (- (* 16 (fact 3)) (fact 3)))) ;; the starting digits 1/A
 (+ 4))                                 ;; add in the next lower count (3 digits)
;; => 262


(defun count-hex-combinations-1 (acc ndigits)
  "This one dind't work out, generalizing from 4 wasn't enough"
  (cond
    ((< ndigits 3)
     (format t "terminating at ndigits<3:~a~&" acc)
     acc)
    ((= ndigits 3)
     (+ acc 4))
    (t
     (count-hex-combinations
      (+ acc
	 (+
	  (* 13 (fact (- ndigits 1)))
	  (* 2 (- (* 16 (fact (- ndigits 1))) (fact (- ndigits 1))))))
      (- ndigits 1)))))

;; length(1):0
;; length(2):0
;; length(3):4
;; length(4):262
;; length(5):10190

;(count-hex-combinations-1 0 4) ; expect 262, got 262 (good)
;(count-hex-combinations-1 0 5) ; expect 10190, got 1294 (no good)
; (format t "~x" (count-hex-combinations-1 0 16)) ; 36D0732C01B6

;; when len = 5, and the first char is one of (2,3,4,5,6,7,8,9,B,C,D,E,F)
;;     2*01A 13 * 16 * fact(3)

(defun count-hex-combinations-2 (acc ndigits)
  (cond
    ((< ndigits 3)
     (format t "terminating at ndigits<3:~a~&" acc)
     acc)
    ((= ndigits 3)
     (+ acc 4))
    (t
     (count-hex-combinations
      (+ acc
	 (+
	  (* 13 (expt 16 (- ndigits 4)) (fact (- ndigits 1)))
	  (* 2 (- (* 16 (fact (- ndigits 1))) (fact (- ndigits 1))))))
      (- ndigits 1)))))

;(count-hex-combinations-2 0 4) ; expect 262, got 262 (good)
;(count-hex-combinations-2 0 5) ; expect 10190, got 5974 (no good)

;; when len = 5, and the first char is one of (2,3,4,5,6,7,8,9,B,C,D,E,F)
;;     .*01A 13 * 16 * fact(3)
;;     A**01
;;     1**0A

(defun count-hex-combinations-3 (acc ndigits)
  (cond
    ((< ndigits 3)
     (format t "terminating at ndigits<3:~a~&" acc)
     acc)
    ((= ndigits 3)
     (+ acc 4))
    (t
     (count-hex-combinations
      (+ acc
	 (+
	  (* 13 (expt 16 (- ndigits 4)) (fact (- ndigits 1)))
	  (* 2 (- (* 16 (fact (- ndigits 1))) (fact (- ndigits 1))))))
      (- ndigits 1)))))
